/*
	LineUpScenes plugin for AviSynth

	Copyright (C) 2006 Mohan
  
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.
	 
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
		
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	
	The author can be contacted at the doom9.org forum under mg262.
*/

#include "stdafx.h"

#include <vector>
#include <list>
#include <fstream>
#include <iomanip>
using namespace std;

struct Scene
{   
	int begin, end;
	int length() const {return end - begin;}
	Scene (int _begin, int _end) : begin(_begin), end(_end) {}
};

struct ScenePair
{   
	bool xactive, yactive;
	Scene x, y;
	ScenePair(bool _xactive, bool _yactive, Scene _x, Scene _y) : xactive(_xactive), yactive(_yactive), x(_x), y(_y){}
};

class Matcher
{   
public:
	virtual bool valid() = 0;
	virtual int cost() = 0;
	virtual void addresult(list<ScenePair> &result) const = 0;
};
typedef Matcher *PMatcher; 

class Table
{   
public:
	//virtual int value(int x, int y) const = 0;
	virtual PMatcher entry(int x, int y) const = 0;
	virtual Scene xscene(int x) const = 0;
	virtual Scene yscene(int y) const = 0;

	virtual ScenePair pairscenes(Scene, Scene) const = 0;
	virtual ScenePair pairscenes(Scene, bool) const = 0;
	virtual ScenePair pairscenes(bool, Scene) const = 0;
};

class Zero : public Matcher
{   
public:
	virtual bool valid() {return true;}
	virtual int cost() {return 0;}
	virtual void addresult(list<ScenePair> &result) const {return;}
};

class PrecomputedMatcher : public Matcher //implementational convenience
{
protected:
	bool _valid;
	int _cost;
	Matcher *previous;
	list<ScenePair> pairs;
public:
	virtual bool valid() {return _valid;}
	virtual int cost() {return _cost;}
	virtual void addresult(list<ScenePair> &result) const;
};

void PrecomputedMatcher::addresult(list<ScenePair> &result) const
{
	previous->addresult(result);
	result.insert(result.end(), pairs.begin(), pairs.end());
}

class Pair : public PrecomputedMatcher
{  
public:
	Pair(Table &, int x, int y);
};
	
Pair::Pair(Table &table, int x, int y) //: table(_table), x(_x), y(_y)
{   
	_valid = (x > 0 && y > 0);
	if (_valid)
	{
		previous = table.entry(x-1,y-1);
		Scene sx = table.xscene(x);
		Scene sy = table.yscene(y);

		int difference = sx.length() -  sy.length();
		_cost = previous->cost() + difference*difference;
		pairs.push_back(table.pairscenes(sx, sy));
	}
}

class Skip : public PrecomputedMatcher
{   
//	const static int skipcost = 40;
public:
	Skip(Table &, int x, int y, int skipcost);
};
	
Skip::Skip(Table &table, int x, int y, int skipcost)
{   
	_valid = (x > 0);
	if (_valid)
	{
		previous = table.entry(x-1,y);
		Scene sx = table.xscene(x);

		_cost = previous->cost() + skipcost;
		pairs.push_back(table.pairscenes(sx, false));
	}
}

class Merge : public PrecomputedMatcher
{   
public:
	Merge(Table &, int x, int y, int mergecost);
};
	
Merge::Merge(Table &table, int x, int y, int mergecost)
{   
	_valid = (x >= 2 && y >= 1);
	if (_valid)
	{
		previous = table.entry(x-2,y-1);
		Scene sx0 = table.xscene(x-1);
		Scene sx1 = table.xscene(x);
		Scene merged(sx0.begin, sx1.end);
		Scene sy = table.yscene(y);

		int difference = merged.length() -  sy.length();
		_cost = previous->cost() + difference*difference + mergecost;
		pairs.push_back(table.pairscenes(merged, sy));
	}
}
/*class TransposeTable : public Table
{   
	Table &table;
public:
	TransposeTable(Table &_table) : table (_table) {}
	virtual int value(int x, int y) const {return table.value(y, x);}
	virtual Scene xscene(int x) const {return table.yscene(x);}
	virtual Scene yscene(int y) const {return table.xscene(y);}
};/*/

class TransposeTable : public Table
{   
	const Table *table;
public:
	TransposeTable(Table *_table) : table (_table) {}
	void settable(Table *_table) {table = _table;}
	//virtual int value(int x, int y) const {return table->value(y, x);}
	virtual PMatcher entry(int x, int y) const {return table->entry(y, x);}
	virtual Scene xscene(int x) const {return table->yscene(x);}
	virtual Scene yscene(int y) const {return table->xscene(y);}

	virtual ScenePair pairscenes(Scene _x, Scene _y) const{return ScenePair(true, true, _y, _x);}
	virtual ScenePair pairscenes(Scene _x, bool) const{return ScenePair(false, true, Scene(0,0), _x);}
	virtual ScenePair pairscenes(bool, Scene _y) const{return ScenePair(true, false, _y, Scene(0,0));}
};

class CostTable : public Table
{   
	vector<Scene> xscenes, yscenes;
	int nx, ny;
	const int skipcost;
	const int mergecost;

	vector<Matcher *> data;
	TransposeTable transpose;

	PMatcher &entry(int x, int y) {return data.at(x+nx*y);}
	//PMatcher entry(int x, int y) const {return data.at(x+nx*y);}
	void update(int x, int y, Matcher *matcher);
public:
	CostTable(vector<Scene> _xscenes, vector<Scene> _yscenes, int _skipcost, int _mergecost);
	//virtual int value(int x, int y) const {return entry(x, y)->cost();}
	virtual PMatcher entry(int x, int y) const {return data.at(x+nx*y);}
	virtual Scene xscene(int x) const {return xscenes.at(x);}
	virtual Scene yscene(int y) const {return yscenes.at(y);}

	virtual ScenePair pairscenes(Scene _x, Scene _y) const{return ScenePair(true, true, _x, _y);}
	virtual ScenePair pairscenes(Scene _x, bool) const{return ScenePair(true, false, _x, Scene(0,0));}
	virtual ScenePair pairscenes(bool, Scene _y) const{return ScenePair(false, true, Scene(0,0), _y);}

	list<ScenePair> optimal() const;
};

CostTable::CostTable(vector<Scene> _xscenes, vector<Scene> _yscenes, int _skipcost, int _mergecost)
: transpose(0), xscenes(_xscenes), yscenes(_yscenes), nx(int(_xscenes.size())), ny(int(_yscenes.size())), data(nx*ny), 
skipcost(_skipcost), mergecost(_mergecost)
{
	transpose.settable(this);
	entry(0, 0) = new Zero;
	for(int y = 0; y < ny; ++y)
		for(int x = 0; x < nx; ++x)
		{
			update(x, y, new Pair(*this, x, y) ); 
			update(x, y, new Skip(*this, x, y, skipcost) );			//skip on the x side
			update(x, y, new Skip(transpose, y, x, skipcost) ); 	//skip on the y side
			update(x, y, new Merge(*this, x, y, mergecost) );			//skip on the x side
			update(x, y, new Merge(transpose, y, x, mergecost) ); 	//skip on the y side
		}
}

void CostTable::update(int x, int y, Matcher *matcher)
{   
	if (!matcher->valid())
		return;

	if (entry(x, y) == 0)
		entry(x, y) = matcher;
	else if (/*entry(x, y) exists and*/ matcher->cost() < entry(x, y)->cost())
	{
		delete entry(x, y);
		entry(x, y) = matcher;
	}
}

list<ScenePair> CostTable::optimal() const
{   
	list<ScenePair> result;
	entry(nx-1, ny-1)->addresult(result);
	return result;
}


vector<Scene> readscenes(const char *filename)
{   
	ifstream input(filename);
	vector<Scene> result;

	int noframes;
	input >> noframes;
	//result.reserve(noframes);

	int begin, end;
	input >> begin >> ws;

	//in order that the n-th row/column of the table can refer to the n-th scene, we need a null scene
	result.push_back(Scene(0,0));
	while (input)
	{
		input >> end;
		result.push_back(Scene(begin, end));
		begin = end;
	}
	result.push_back(Scene(begin, noframes));

	return result;
}

class LineUpScenes : public GenericVideoFilter 
{   
	const static int framesperscene = 4;
	CostTable table;
	PClip x, y;

	vector<ScenePair> pairs;
	PVideoFrame blackframe;
public:
	LineUpScenes(vector<Scene> _xscenes, vector<Scene> _yscenes, PClip _x, PClip _y, 
		const char *filename, int skipcost, int mergecost, IScriptEnvironment* env);
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};

void blackenplane(PVideoFrame &frame, int plane)
{
	const int width = frame->GetRowSize(plane);	
	const int height = frame->GetHeight(plane);	
	const int pitch = frame->GetPitch(plane);	
	const int colour = (plane==PLANAR_Y) ? 0 : 128;

	unsigned char* row = frame->GetWritePtr(plane);
	for (int y = 0; y < height; ++y) 
	{
		for (int x = 0; x < width; ++x)
			row[x] = colour;
		row += pitch;
	}
}

LineUpScenes::LineUpScenes(vector<Scene> _xscenes, vector<Scene> _yscenes, PClip _x, PClip _y, 
						   const char *filename, int skipcost, int mergecost, IScriptEnvironment* env)
:GenericVideoFilter(_x), table(_xscenes, _yscenes, skipcost, mergecost), x(_x), y(_y)
{
	blackframe = env->NewVideoFrame(vi);
	blackenplane(blackframe, PLANAR_Y);
	blackenplane(blackframe, PLANAR_U);
	blackenplane(blackframe, PLANAR_V);

	const list<ScenePair> linedup = table.optimal();
	pairs.reserve(linedup.size());
	for(list<ScenePair>::const_iterator i = linedup.begin(); i != linedup.end(); ++i)
		pairs.push_back(*i);
	//copy(linedup.begin(), linedup.end(), back_inserter(pairs));

	vi.num_frames = int(pairs.size()) * (2 * framesperscene);

	ofstream output(filename);
	for(vector<ScenePair>::iterator i = pairs.begin(); i != pairs.end(); ++i)
	{
		if (i->xactive)
			output << i->x.begin << "..." << i->x.end-1;
		output << "\t";
		if (i->yactive)
			output << i->y.begin << "..." << i->y.end-1;
		output << "\t";
		if (i->xactive && i->yactive)
			output << "" << i->y.end-i->x.end;
		output << "\n";
	}
}

int selectspaced(Scene s, int numerator, int denominator)
{
	return s.begin + ((s.end - s.begin - 1)*numerator)/denominator;
}

PVideoFrame __stdcall LineUpScenes::GetFrame(int n, IScriptEnvironment* env) 
{
	if (!vi.IsYV12()) 
		env->ThrowError("LineUpScenes requires YV12.");

	int sceneno = n/(2 * framesperscene);
	int horizontal = n%(2 * framesperscene);
	int clipno = horizontal/framesperscene;
	int index = horizontal%framesperscene;

	ScenePair pair = pairs[sceneno];
	if (clipno == 0)
		return pair.xactive ? x->GetFrame(selectspaced(pair.x, index, framesperscene-1), env) : blackframe;
	else
		return pair.yactive ? y->GetFrame(selectspaced(pair.y, index, framesperscene-1), env) : blackframe;
}



AVSValue __cdecl Create_LineUpScenes(AVSValue args, void* user_data, IScriptEnvironment* env) 
{
	return new LineUpScenes(
		readscenes(args[0].AsString()), 
		readscenes(args[1].AsString()), 
		args[2].AsClip(), 
		args[3].AsClip(), 
		args[4].AsString("lineup.txt"),
		args[5].AsInt(20), 
		args[6].AsInt(25), 
		env);  
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) 
{
	env->AddFunction("LineUpScenes", "sscc[text]s[skipcost]i[mergecost]i", Create_LineUpScenes, 0);
	return "LineUpScenes plugin";
}
